// -*- Mode: c++ -*-
// $Id: subdivide.h,v 1.3 2000/07/19 19:08:15 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/include/subdivide.h,v $

/* Subdivide V2.0
   Copyright (C) 2000 Henning Biermann, Denis Zorin, NYU

This file is part of Subdivide.

Subdivide is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Subdivide is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Subdivide; see the file COPYING.  If not, write to the Free
Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef __SUBDIVIDE_H__
#define __SUBDIVIDE_H__

#include "compat.h"



// Subdivider: a set of functions to apply subdivision rules
// to a ring

template<class RuleTable, class FaceRing>
class SubdivideTp {
public:

  typedef typename FaceRing::FaceType FaceType;

  static bool hdRelevantToMidpoint(FaceType* f, EnoType e);
  void getHdRuleAndRing(FaceType* f, EnoType e, int depth);

  cvec3d computeVertexPoint() const;
  cvec3d computeEdgePoint() const;
  cvec3d computeNormal() const;
  cvec3d computeLimit() const;

  const typename RuleTable::RuleType* rule() { return _rule; }
  const FaceRing& ring() const { return _faceRing; }
  cvec3d modifyFlatness(const cvec3d& p, double x1, double x2) const;
  cvec3d modifyClosedNormal(const cvec3d& p, double x1, double x2) const;
  cvec3d modifyOpenNormal(const cvec3d& p, double x1, double x2) const;
  void printRelevance() const;

private:
  int _depth;
  typename RuleTable::RuleType* _rule;
  FaceRing _faceRing;
};

#include "subdivide.hi"

#endif /* __SUBDIVIDE_H__ */
