// -*- Mode: c++ -*-
// $Id: silviewer.cpp,v 1.21 2000/08/09 19:45:51 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/silviewer.cpp,v $

#include <string.h>
#include <GL/glut.h>
#include "tagivgraph.h"
#include "tagflatmesh.h"
#include "options.h"
#include "silhouette.h"
#include "dualviewer.h"
#include "meshviewer.h"
#include "silviewer.h"

SilViewer::SilViewer(char *title, int width, int height)
  : Viewer(title, width, height),
    _isRendering(true),
    _zooming(false), _centerX(width/2.0), _centerY(height/2.0), _zoom(1.0),
    _translating(false), _transX(0), _transY(0),
    _mesh(0), _min(0), _max(0), _meshviewer(0), _grid(), _uiAction(0),
    _camera(), _arcball(),
    _dualmesh(0), _dualviewer(0), _silhouettes(0) {
  glDisable(GL_DEPTH_TEST);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, _width, 0.0, _height, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

SilViewer::~SilViewer() {
  if (_uiAction) delete _uiAction;
  _uiAction = 0;
  if (_dualviewer) delete _dualviewer;
  _dualviewer = 0;
  if (_meshviewer) delete _meshviewer;
  _meshviewer = 0;
  if (_silhouettes) delete _silhouettes;
  _silhouettes = 0;
  if (_dualmesh) delete[] _dualmesh;
  _dualmesh = 0;
  if (_mesh) delete _mesh;
  _mesh = 0;
}

bool SilViewer::readMesh(char *meshfile) {
  // read the wrl file
  TagIvGraph tagIvGraph;
  if (! tagIvGraph.read(meshfile)) return false;
 
  // convert from wrl to flat mesh
  // second parameter (true) indicates to triangulate all faces
  TagFlatMesh tagFlatMesh;
  if (! tagIvGraph.toTagFlatMesh(&tagFlatMesh, true)) return false;
 
  // convert from flat mesh to tri mesh
  if (_mesh) delete _mesh;
  _mesh = new TMesh(tagFlatMesh);

  return true;
}

void SilViewer::subdivide(int d) {
  if (_mesh) {
    cerr << "subdivide to depth " << d << " ... ";
    _mesh->subdivide(d);
    _grid.placeIntoGrid(_mesh);
    cerr << "done." << endl;
  }
}

void SilViewer::setupViewer() {
  // compute min & max
  TMesh::FaceIterType ti = _mesh->faceBegin();
  _max = _min = (*ti)->vert(0)->limitPos();
  for(; ti != _mesh->faceEnd(); ++ti) {
    if((*ti)->isLeaf()) {
      for(VnoType v = 0; v < (*ti)->noVtx(); ++v) {
	cvec3d p = (*ti)->vert(v)->limitPos();
	_max = _max.max(p);
	_min = _min.min(p);
      }
    }
  }
  positionCamera();
  centerArcBall();
}

void SilViewer::outputCamera(ostream &c) {
  c << _camera;
}

void SilViewer::inputCamera(istream &c) {
  c >> _camera;
  recomputeSilhouettes();
  redisplay();
  if (_dualviewer) _dualviewer->redisplay();
  if (_meshviewer) _meshviewer->redisplay();
}

void SilViewer::redisplay() {
  _isRendering = true;
  glutSetWindow(_winId);
  glutPostRedisplay();
}

void SilViewer::initSilhouettes() {
  if (! _mesh) return;

  // build dual mesh
  if (_dualmesh) delete[] _dualmesh;
  cerr << "building dual mesh ... ";
  _dualmesh = OcTree::buildTree(_mesh);
  if (_dualviewer) _dualviewer->setDualMesh(_dualmesh);
  cerr << "done." << endl;

  if (!_silhouettes) _silhouettes = new Silhouettes(grid(), &_camera);
  _silhouettes->clear();
  _silhouettes->createGrid(_grid.size());

  // find silhouettes
  cerr << "looking for silhouettes ... ";
  cvec3d pos = _camera.viewWorldPosition();
  if (_dualviewer) _dualviewer->setCameraPosition(pos);

  _dualmesh[0].findSilhouettes(cvec3d(pos.y(), pos.z(), 1), pos.x(),
			       _silhouettes);
  _dualmesh[1].findSilhouettes(cvec3d(pos.z(), 1, pos.x()), pos.y(),
			       _silhouettes);
  _dualmesh[2].findSilhouettes(cvec3d(1, pos.x(), pos.y()), pos.z(),
			       _silhouettes);
  _dualmesh[3].findSilhouettes(pos, 1, _silhouettes);
  _dualmesh[4].findSilhouettes(cvec3d(pos.y(), pos.z(), 1), -pos.x(),
			       _silhouettes);
  _dualmesh[5].findSilhouettes(cvec3d(pos.z(), 1, pos.x()), -pos.y(),
			       _silhouettes);
  _dualmesh[6].findSilhouettes(cvec3d(1, pos.x(), pos.y()), -pos.z(),
			       _silhouettes);
  _dualmesh[7].findSilhouettes(pos, -1, _silhouettes);
  cerr << "done." << endl;
  DualEdge::clearCheckTags();

  // insert boundaries
  if (Options::takeBoundary) {
    int bsz = OcTree::boundary.size();
    for (int i = 0; i < bsz; i++)
      _silhouettes->insertEdge(&(OcTree::boundary[i]));
  }

  // create chains
  if (Options::visibility) {
    cerr << "create silhouette chains ... ";
    _silhouettes->createChains();
    cerr << "done." << endl;
  }
}

void SilViewer::recomputeSilhouettes() {
  _silhouettes->clear();

  cvec3d pos = _camera.viewWorldPosition();
  if (_dualviewer) _dualviewer->setCameraPosition(pos);

  _dualmesh[0].findSilhouettes(cvec3d(pos.y(), pos.z(), 1), pos.x(),
			       _silhouettes);
  _dualmesh[1].findSilhouettes(cvec3d(pos.z(), 1, pos.x()), pos.y(),
			       _silhouettes);
  _dualmesh[2].findSilhouettes(cvec3d(1, pos.x(), pos.y()), pos.z(),
			       _silhouettes);
  _dualmesh[3].findSilhouettes(pos, 1, _silhouettes);
  _dualmesh[4].findSilhouettes(cvec3d(pos.y(), pos.z(), 1), -pos.x(),
			       _silhouettes);
  _dualmesh[5].findSilhouettes(cvec3d(pos.z(), 1, pos.x()), -pos.y(),
			       _silhouettes);
  _dualmesh[6].findSilhouettes(cvec3d(1, pos.x(), pos.y()), -pos.z(),
			       _silhouettes);
  _dualmesh[7].findSilhouettes(pos, -1, _silhouettes);

  DualEdge::clearCheckTags();

  // insert boundaries
  if (Options::takeBoundary) {
    int bsz = OcTree::boundary.size();
    for (int i = 0; i < bsz; i++)
      _silhouettes->insertEdge(&(OcTree::boundary[i]));
  }

  if (Options::visibility) _silhouettes->createChains();
}

void SilViewer::positionCamera() {
  if (!_mesh) return;

  int vp[] = {0, 0, _width, _height};
  _camera.reset();
  _camera.setViewport(vp);

  cvec3d mid = 0.5 * (_min + _max);
  
  double w = max(_max.x() - _min.x(), 
                _max.y() - _min.y()) /
    2.0 / tan(0.5 * _camera.fovy() * M_PI / 180.0);
  
  cvec3d trans(mid.x(), mid.y(), _max.z() + 2 * w);

  double d = 4 * w + 3 * max( w, _max.z() - _min.z() );

  _camera.setPerspectiveParams(_camera.fovy(), 0.1, d );

  _camera.translate(-trans);
  _camera.computeModelview();
  _camera.computeProjection();
  // _camera.loadMatrices();
}

void SilViewer::centerArcBall() {
  double z = _camera.project(0.5 * (_min + _max)).z();
  
  double top  =  _height * 5.0/6.0;
  double right = _width  * 5.0/6.0;
  double cx = _width * 0.5;
  double cy = _height * 0.5;
  
  cvec3d pos = _camera.unproject(cvec3d(cx, cy, z));
  double rad = min((_camera.unproject(cvec3d(right, cy, z)) - pos).l2(),
                  (_camera.unproject(cvec3d(cx, top, z)) - pos).l2());

  _arcball.Init();
  _arcball.SetPosition(pos);
  _arcball.SetRadius(rad);
  _arcball.Update();
}

void SilViewer::display() {
  if (Options::visibility) render();
  else renderRaw();
}

void SilViewer::reshape(int w, int h) {
  if ((w == 0) || (h == 0) || ((w == _width) && (h == _height))) return;
  _centerX *= double(w)/_width;
  _centerY *= double(h)/_height;
  _transX *= double(w)/_width;
  _transY *= double(h)/_height;
  _width = w;
  _height = h;
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(_centerX + _transX*_zoom - _width/2*_zoom,
	  _centerX + _transX*_zoom + _width/2*_zoom,
	  _centerY + _transY*_zoom - _height/2*_zoom,
	  _centerY + _transY*_zoom + _height/2*_zoom,
	  -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);

  int vp[] = {0, 0, w, h};
  _camera.setViewport(vp);
  _camera.computeProjection();
  recomputeSilhouettes();
  redisplay();
}

void SilViewer::mouse(int button, int state, int x, int y) {
  _zooming = false;
  _translating = false;
  if (_uiAction) {
    delete _uiAction;
    _uiAction = 0;
  }
  y = _height - y; 

  if (state == GLUT_DOWN) {
    switch(button) {
    case GLUT_LEFT_BUTTON: 
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
	_translating = true;
	_startX = x, _startY = y;
      } else {
	_uiAction = new CameraRotateAction(x, y, &_arcball, &_camera); 
      }
      break;
    case GLUT_MIDDLE_BUTTON:
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
	_uiAction = new CameraTransXYAction(x, y, &_arcball, &_camera);
      }
      // translate, not touching camera
      else {
	_translating = true;
	_startX = x, _startY = y;
      }
      break;
    case GLUT_RIGHT_BUTTON:
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
	_uiAction = new CameraTransZAction(x, y, &_arcball, &_camera);
      }
      // zoom, not touching camera
      else {
	_zooming = true;
	_zoomY = y;
	_start = _zoom;
	_centerX += _transX*_zoom;
	_centerY += _transY*_zoom;
	_transX = 0;
	_transY = 0;
      }
      break;
    }

    if (_uiAction && _meshviewer) {
      _meshviewer->camera().copyModelOnly(_camera);
      _meshviewer->redisplay();
    }
  }
  redisplay();
}

void SilViewer::motion(int x, int y) {
  if (isRendering()) return;

  y = _height - y; 
  if (_zooming) {
    _zoom = _start*pow(1.01, double(_zoomY - y));

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(_centerX + _transX*_zoom - _width/2*_zoom,
	    _centerX + _transX*_zoom + _width/2*_zoom,
	    _centerY + _transY*_zoom - _height/2*_zoom,
	    _centerY + _transY*_zoom + _height/2*_zoom,
	    -1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);

    redisplay();
  } else if (_translating) {
    _transX += (_startX - x);
    _transY += (_startY - y);
    _startX = x;
    _startY = y;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(_centerX + _transX*_zoom - _width/2*_zoom,
	    _centerX + _transX*_zoom + _width/2*_zoom,
	    _centerY + _transY*_zoom - _height/2*_zoom,
	    _centerY + _transY*_zoom + _height/2*_zoom,
	    -1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);

    redisplay();
  } else if (_uiAction) {
    _uiAction->update(x, y);
    recomputeSilhouettes();
    redisplay();
    if (_dualviewer) _dualviewer->redisplay();
    if (_meshviewer) {
      _meshviewer->camera().copyModelOnly(_camera);
      _meshviewer->redisplay();
    }
  }
}

void SilViewer::renderRaw() {
  // render the silhouettes directly, for debugging
  glLineWidth(2);
  glBegin(GL_LINES);
  int esz = _silhouettes->edges().size();
  for (int i=0; i<esz; i++) {
    if (_silhouettes->edges()[i]->isBoundary()) glColor3ub(0, 255, 0);
    else glColor3ub(0, 0, 255);
    glVertex3dv(_silhouettes->edges()[i]->iv(0));
    glVertex3dv(_silhouettes->edges()[i]->iv(1));
  }
  glEnd();

  _isRendering = false;
}

void SilViewer::render() {
  if (!_silhouettes) return;

  GLubyte color[][3] = {{255, 0, 0}, {0, 255, 0}, {0, 0, 255},
			{255, 0, 255}, {0, 255, 255}, {0, 0, 0}};
  int ci = 0, csz = _silhouettes->chains().size();

  for (int it = 0; it < csz; it++) {
    Chain *chain = _silhouettes->chains()[it];
    if (chain->visible() || Options::drawHiddenCurves) {
      // select a color
      if (Options::distinctChainColors) {
	glColor3ubv(color[ci]);
	ci = (ci+1) % 6;
      } else {
	if (chain->visible()) glColor3ub(0, 0, 0);
	else glColor3ub(180, 180, 180);
      }

      // render the chain
      glLineWidth(2);
      glBegin(GL_LINE_STRIP);
      int psz = chain->points().size();
      for (int i = 0; i < psz; i++) {
	glVertex2dv(chain->points()[i]);
      }
      glEnd();

      // draw QI
      if (Options::drawQI) {
	cvec3d pos = chain->points()[0].lerp(chain->points()[1], 0.2);
	glPushMatrix();
	glTranslated(pos.x(), pos.y(), 0);
	glScaled(0.1*_zoom, 0.1*_zoom, 0.1*_zoom);
	glLineWidth(1);
	glColor3ub(0, 0, 255);
	glutStrokeCharacter(GLUT_STROKE_ROMAN, chain->QI() + int('0'));
	glPopMatrix();
      }

      // draw intersections
      if (Options::drawIntersections) {
	glPointSize(3);
	glColor3ub(255, 0, 0);
	glBegin(GL_POINTS);
	{
	  SilEdge *se = chain->edge(0);
	  int isz = se->nInter();
	  for (int k = 0; k < isz; k++) {
	    glVertex2dv(se->inter_pos(k));
	  }
	}
	int s = chain->size();
	if (s>1) {
	  SilEdge *se = chain->edge(s-1);
	  int isz = se->nInter();
	  for (int k = 0; k < isz; k++) {
	    glVertex2dv(se->inter_pos(k));
	  }
	}
	glEnd();
      }
    }
  }

  // draw grid
  if (Options::drawSilGrid && _silhouettes->grid()) {
    cvec3d maxp = _silhouettes->grid()->maxp(),
      minp = _silhouettes->grid()->minp(), s = maxp - minp;
    double xstep = _silhouettes->grid()->xstep(),
      ystep = _silhouettes->grid()->ystep();

    glLineWidth(1);
    glColor3ub(0, 200, 0);
    glBegin(GL_LINES);
    int gsz = _silhouettes->grid()->size();
    for (int i=0; i<=gsz; i++) {
      glVertex2dv(minp + cvec3d(0,     i*ystep, 0));
      glVertex2dv(minp + cvec3d(s.x(), i*ystep, 0));
      glVertex2dv(minp + cvec3d(i*xstep, 0,     0));
      glVertex2dv(minp + cvec3d(i*xstep, s.y(), 0));
    }
    glEnd();
  }

  _isRendering = false;
}

void SilViewer::viewMesh(bool view) {
  if (view) {
    if (!_meshviewer) {
      _meshviewer = new
	MeshViewer(_camera, &_arcball, &_grid,
		   _silhouettes, _width, _height);
      _meshviewer->setMesh(_mesh);
    }
  } else {
    if (_meshviewer) delete _meshviewer;
    _meshviewer = 0;
  }
}

void SilViewer::viewDualMesh(bool view) {
  if (view) {
    if (!_dualviewer) {
      _dualviewer = new
	DualViewer(Options::dualWinSize, Options::dualWinSize);
      _dualviewer->setDualMesh(_dualmesh);
    }
  } else {
    if (_dualviewer) delete _dualviewer;
    _dualviewer = 0;
  }
}

void SilViewer::useMeshCamera()
{ if (_meshviewer) _camera.copyModelOnly(_meshviewer->camera()); }

void SilViewer::clearMeshDisplayList()
{ if (_meshviewer) _meshviewer->clearDisplayList(); }

void SilViewer::redisplayMesh()
{ if (_meshviewer) _meshviewer->redisplay(); }

void SilViewer::redisplayDualMesh()
{ if (_dualviewer) _dualviewer->redisplay(); }
