// -*- mode: c++ -*-
// $Id: silviewer.h,v 1.11 2000/07/23 19:54:17 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/silviewer.h,v $


#ifndef __SILVIEWER_H__
#define __SILVIEWER_H__

#include <GL/glut.h>
#include "camera.h"
#include "arcball.h"
#include "uiaction.h"
#include "subtri.h"
#include "tagmesh.h"
#include "grid3d.h"
#include "viewer.h"

class Silhouettes;
class OcTree;
class DualViewer;
class MeshViewer;

class SilViewer : public Viewer {

public:
  SilViewer(char *title = 0, int width = 400, int height = 300);
  virtual ~SilViewer();

  // rendering flag
  bool isRendering() { return _isRendering; };

  // original mesh
  TMesh* mesh() { return _mesh; };

  // 3D grid
  Grid3D* grid() { return &_grid; };

  bool readMesh(char *meshfile);
  void subdivide(int d);
  void setupViewer();

  // camera
  Camera& camera() { return _camera; };

  // copy camera from mesh viewer
  void useMeshCamera();

  // output and input camera
  void outputCamera(ostream &c);
  void inputCamera(istream &c);

  // arcball
  ArcBall& arcball() { return _arcball; };

  // dual mesh
  OcTree* dualmesh() { return _dualmesh; };

  // viewed silhouettes
  Silhouettes* silhouettes() { return _silhouettes; };

  // redisplay window
  void redisplay();

  // original mesh viewer
  void viewMesh(bool view);
  void clearMeshDisplayList();
  void redisplayMesh();

  // dual mesh viewer
  void viewDualMesh(bool view);
  void redisplayDualMesh();

  // find silhouettes
  void initSilhouettes();
  void recomputeSilhouettes();

private:
  // rendering flag
  bool _isRendering;

  // zoom (not changing camera)
  bool _zooming; // flag
  int _zoomY;    // start y
  double _centerX, _centerY;
  double _start, _zoom;  // zoom factor

  // translate (not changing camera)
  bool _translating;      // flag
  int _startX, _startY;
  double _transX, _transY;

  // original mesh
  TMesh *_mesh;
  cvec3d _min, _max;
  MeshViewer *_meshviewer;

  // 3D grid
  Grid3D _grid;

  UiAction *_uiAction;

  // camera
  Camera _camera;

  // arcball
  ArcBall _arcball;

  // dual mesh
  OcTree *_dualmesh;
  DualViewer *_dualviewer;

  // silhouettes to view
  Silhouettes *_silhouettes;

  void render();
  void renderRaw();

  void positionCamera();
  void centerArcBall();

  void display();
  void reshape(int w, int h);
  void mouse(int button, int state, int x, int y);
  void motion(int x, int y);
};

#endif /* __SILVIEWER_H__ */
